@@ -191,6 +191,7 @@ struct _ChamplainViewPrivate
   gint tiles_loading;
   
   guint redraw_timeout;
+  guint zoom_timeout;
   
   ClutterAnimationMode goto_mode;
   guint goto_duration;
@@ -214,6 +215,7 @@ struct _ChamplainViewPrivate
   gdouble focus_lat;
   gdouble focus_lon;
   gboolean zoom_started;
+  gdouble accumulated_scroll_dy;
 
   ChamplainBoundingBox *world_bbox;
 };
@@ -275,6 +277,9 @@ static void get_tile_bounds (ChamplainView *view,
     guint *min_y,
     guint *max_x,
     guint *max_y);
+static gboolean tile_in_tile_map (ChamplainView *view,
+    gint tile_x,
+    gint tile_y);
 
 static void
 update_coords (ChamplainView *view,
@@ -385,6 +390,21 @@ panning_completed (G_GNUC_UNUSED ChamplainKineticScrollView *scroll,
 }
 
 
+static gboolean
+zoom_timeout_cb (gpointer data)
+{
+  DEBUG_LOG ()
+
+  ChamplainView *view = data;
+  ChamplainViewPrivate *priv = view->priv;
+
+  priv->accumulated_scroll_dy = 0;
+  priv->zoom_timeout = 0;
+
+  return FALSE;
+}
+
+
 static gboolean
 scroll_event (G_GNUC_UNUSED ClutterActor *actor,
     ClutterScrollEvent *event,
@@ -403,18 +423,23 @@ scroll_event (G_GNUC_UNUSED ClutterActor *actor,
   else if (event->direction == CLUTTER_SCROLL_SMOOTH)
     {
       gdouble dx, dy;
+      gint steps;
 
       clutter_event_get_scroll_delta ((ClutterEvent *)event, &dx, &dy);
 
-      /* Use a threshhold value to avoid jitter */
-      if (fabs(dy) < 0.75)
-        return FALSE;
-
+      priv->accumulated_scroll_dy += dy;
+      /* add some small value to avoid missing step for values like 0.999999 */
       if (dy > 0)
-        zoom_level = priv->zoom_level - 1;
+        steps = (int) (priv->accumulated_scroll_dy + 0.01);
       else
-        zoom_level = priv->zoom_level + 1;
-  }
+        steps = (int) (priv->accumulated_scroll_dy - 0.01);
+      zoom_level = priv->zoom_level - steps;
+      priv->accumulated_scroll_dy -= steps;
+
+      if (priv->zoom_timeout != 0)
+        g_source_remove (priv->zoom_timeout);
+      priv->zoom_timeout = g_timeout_add (1000, zoom_timeout_cb, view);
+    }
 
   return view_set_zoom_level_at (view, zoom_level, TRUE, event->x, event->y);
 }
@@ -673,12 +698,18 @@ champlain_view_dispose (GObject *object)
       priv->redraw_timeout = 0;
     }
     
-  if (priv->redraw_timeout != 0)
+  if (priv->zoom_actor_timeout != 0)
     {
       g_source_remove (priv->zoom_actor_timeout);
       priv->zoom_actor_timeout = 0;
     }
 
+  if (priv->zoom_timeout != 0)
+    {
+      g_source_remove (priv->zoom_timeout);
+      priv->zoom_timeout = 0;
+    }
+
   if (priv->tile_map != NULL)
     {
       g_hash_table_destroy (priv->tile_map);
@@ -1248,6 +1279,18 @@ zoom_gesture_finish_cb (ClutterGestureAction *gesture,
 }
 
 
+static void
+zoom_gesture_cancel_cb (ClutterGestureAction *gesture,
+    G_GNUC_UNUSED ClutterActor *actor,
+    gpointer user_data)
+{
+  ChamplainViewPrivate *priv = GET_PRIVATE (user_data);
+
+  priv->zoom_started = FALSE;
+  g_signal_stop_emission_by_name (gesture, "gesture-cancel");
+}
+
+
 static void
 champlain_view_init (ChamplainView *view)
 {
@@ -1355,7 +1398,7 @@ champlain_view_init (ChamplainView *view)
   g_signal_connect (priv->zoom_gesture, "gesture-end",
                     G_CALLBACK (zoom_gesture_finish_cb), view);
   g_signal_connect (priv->zoom_gesture, "gesture-cancel",
-                    G_CALLBACK (zoom_gesture_finish_cb), view);
+                    G_CALLBACK (zoom_gesture_cancel_cb), view);
   clutter_actor_add_action (CLUTTER_ACTOR (view),
                             CLUTTER_ACTION (priv->zoom_gesture));
 
@@ -1743,6 +1786,109 @@ champlain_view_zoom_out (ChamplainView *view)
 }
 
 
+static void
+layers_to_surface (ChamplainView *view,
+    cairo_t *cr)
+{
+  ClutterActorIter iter;
+  ClutterActor *child;
+
+  clutter_actor_iter_init (&iter, view->priv->user_layers);
+  while (clutter_actor_iter_next (&iter, &child))
+    {
+      ChamplainLayer *layer = CHAMPLAIN_LAYER (child);
+      cairo_surface_t *surface;
+
+      if (!CHAMPLAIN_IS_EXPORTABLE (layer))
+        continue;
+
+      surface = champlain_exportable_get_surface (CHAMPLAIN_EXPORTABLE (layer));
+      if (!surface)
+        continue;
+      cairo_set_source_surface (cr, surface, 0, 0);
+      cairo_paint(cr);
+    }
+}
+
+
+/**
+ * champlain_view_to_surface:
+ * @view: a #ChamplainView
+ * @include_layers: Set to %TRUE if you want to include layers
+ *
+ * Will generate a #cairo_surface_t that represents the current view
+ * of the map. Without any markers or layers. If the current #ChamplainRenderer
+ * used does not support this, this function will return %NULL.
+ *
+ * If @include_layers is set to %TRUE all layers that implement
+ * #ChamplainExportable will be included in the surface.
+ *
+ * The #ChamplainView also need to be in #CHAMPLAIN_STATE_DONE state.
+ *
+ * Returns: (transfer full): a #cairo_surface_t or %NULL on failure. Free with
+ *          cairo_surface_destroy() when done.
+ */
+cairo_surface_t *
+champlain_view_to_surface (ChamplainView *view,
+    gboolean include_layers)
+{
+  DEBUG_LOG ()
+
+  g_return_val_if_fail (CHAMPLAIN_IS_VIEW (view), NULL);
+
+  ChamplainViewPrivate *priv = view->priv;
+  cairo_surface_t *surface;
+  cairo_t *cr;
+  ClutterActorIter iter;
+  ClutterActor *child;
+  gdouble width, height;
+
+  if (priv->state != CHAMPLAIN_STATE_DONE)
+    return NULL;
+
+  width = clutter_actor_get_width (CLUTTER_ACTOR (view));
+  height = clutter_actor_get_height (CLUTTER_ACTOR (view));
+  surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, width, height);
+  cr = cairo_create (surface);
+
+  clutter_actor_iter_init (&iter, priv->map_layer);
+  while (clutter_actor_iter_next (&iter, &child))
+    {
+      ChamplainTile *tile = CHAMPLAIN_TILE (child);
+      guint tile_x = champlain_tile_get_x (tile);
+      guint tile_y = champlain_tile_get_y (tile);
+      guint tile_size = champlain_tile_get_size (tile);
+
+      if (tile_in_tile_map (view, tile_x, tile_y))
+        {
+          cairo_surface_t *tile_surface;
+          double x, y, opacity;
+
+          tile_surface = champlain_exportable_get_surface (CHAMPLAIN_EXPORTABLE (tile));
+          if (!tile_surface)
+            {
+              cairo_destroy (cr);
+              cairo_surface_destroy (surface);
+              return NULL;
+            }
+          opacity = ((double) clutter_actor_get_opacity (CLUTTER_ACTOR (tile))) / 255.0;
+          x = ((double) tile_x * tile_size) - priv->viewport_x;
+          y = ((double) tile_y * tile_size) - priv->viewport_y;
+          cairo_set_source_surface (cr,
+                                    tile_surface,
+                                    x, y);
+          cairo_paint_with_alpha(cr, opacity);
+        }
+    }
+
+    if (include_layers)
+      layers_to_surface (view, cr);
+
+    cairo_destroy (cr);
+    return surface;
+}
+
+
 /**
  * champlain_view_set_zoom_level:
  * @view: a #ChamplainView
